/*
    FEARLUS model0-6-2: ObserverSwarm.m
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-2, an agent-based model of land use
    change.

    FEARLUS model0-6-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation for the ObserverSwarm class. As per the header, this is based
heavily on the tutorial material.

*/

#import "ObserverSwarm.h"
#import "ModelSwarm.h"
#import "Environment.h"
#import "LandAllocator.h"
#import "EZBar.h"
#import "DBColormap.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "LandUse.h"
#import "LandParcel.h"
#import "MiscFunc.h"		// getUsableFileNameFrom:withSuffix:
#import "SubPopulation.h"
#import "Reporter.h"
#import "StrategyColourKey.h"
#import <string.h>
#import <math.h>

#define SOLVED_AVERAGE_BUG
#define LPBITSTRINGRASTERTITLE "LP bit "
#define NDIST_SD_MULTIPLIER 2.326341928
				/* When subpopulations use the normal 
				   distribution to initialise certain
				   parameters of the land manager, the
				   graphs showing the distribution should
				   show this many standard deviations
				   about the mean. Roughly 99% should fall
				   in this range. */

@implementation ObserverSwarm

/*

createBegin

Create an ObserverSwarm object. Set up the probes for the class.

*/

+createBegin: (id)z {
  ObserverSwarm *o;
  id <ProbeMap> probeMap;

  o = [super createBegin: z];
  o->displayFrequency = 1;
  o->writeHistoryFrequency = 500;
  o->dumpHistory = YES;
  o->saveDumpedHistory = NO;
  o->historyFileName = (char *)malloc(1024);
  sprintf(o->historyFileName, [FearlusArguments getHistoryFile]);
  o->showStrategyColourKey = 1;
  o->showUseRaster = 1;
  o->showLPBitStringDiffs = 1;
  o->showLPBitStringRaster = 1;
  o->showManagerRaster = 1;
  o->showSubPopRaster = 1;
  o->showStrategyRaster = 1;
  o->showUseChangeRaster = 1;
  o->showManagerChangeRaster = 1;
  o->showParcelsOwnedGraph = 1;
  o->showYieldGraph = 1;
  o->showPopulationGraph = 1;
  o->showUseGraph = 1;
  o->showWealthGraph = 1;
  o->showAgeGraph = 1;
  o->showNWeightDist = 1;
  o->showThresholdDist = 1;
  o->showImitProbDist = 1;
  o->zoomSize = 1;
  o->nHistogramBins = 8;
  o->showUseBar = 0;

  probeMap = [EmptyProbeMap createBegin: z];
  [probeMap setProbedClass: [self class]];
  probeMap = [probeMap createEnd];

  [probeMap addProbe: [probeLibrary 
			getProbeForVariable: "showStrategyColourKey"
			inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showUseRaster"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary
			getProbeForVariable: "showLPBitStringRaster"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showLPBitStringDiffs"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showManagerRaster"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showStrategyRaster"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showUseChangeRaster"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary
			getProbeForVariable: "showManagerChangeRaster"
			inClass: [self class]]];
  [probeMap addProbe: [probeLibrary
			getProbeForVariable: "showParcelsOwnedGraph"
			inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showYieldGraph"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showPopulationGraph"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showUseGraph"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showWealthGraph"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showAgeGraph"
				    inClass: [self class]]];
  /*
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showUseBar"
				    inClass: [self class]]];
  */
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showNWeightDist"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showThresholdDist"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "showImitProbDist"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "nHistogramBins"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "zoomSize"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "displayFrequency"
				    inClass: [self class]]];
  [probeMap addProbe:
	      [probeLibrary getProbeForVariable: "writeHistoryFrequency"
			    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "dumpHistory"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "saveDumpedHistory"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForVariable: "historyFileName"
				    inClass: [self class]]];
  [probeMap addProbe: [probeLibrary getProbeForMessage: "writeHistory"
				    inClass: [self class]]];
  // I might want to set up probes for the suggestion in the header file --
  // i.e. have some flags or options for which displays to use.
  [probeLibrary setProbeMap: probeMap For: [self class]];
  return(o);
}

/*

createEnd

I'm sure this over-riding of the super class createEnd is utterly pointless,
since it doesn't do anything else

*/

-createEnd {
  return([super createEnd]);
}

/*

buildGUIObjects

Build the (non-Swarm standard) GUI objects

*/

-(void)buildGUIObjects {
  //  [globalTkInterp eval:
  //		    "tk_getOpenFile -filetypes {{{LandParcel Files} {.lp}}}"];
  //  printf("%s\n", [globalTkInterp result]);
}

/*

buildReporter

Build the reporter object, and its schedule

*/

-buildReporter {
  reporter = [Reporter createBegin: self];
  [reporter setParameters: parameter];
  [reporter setModelSwarm: modelSwarm];
  reporter = [reporter createEnd];
  reporterSchedule = [Schedule createBegin: self];
  [reporterSchedule setRepeatInterval: 1];
  reporterSchedule = [reporterSchedule createEnd];
  [reporterSchedule at: 0 createActionTo: reporter message: M(report)];
  [reporterSchedule activateIn: self];

  return self;
}

/*

buildObjects

Build the objects needed for the class

*/

-buildObjects {
  double r = 0.0, g = 0.0, b = 0.0;
  int i, j, initZoom, maxEnvDim, nColours, startColour, subpopStartColour;
  id <String> colstr;
  char **colours = NULL;
  char **subpopcolours = NULL;

  [super buildObjects];
  // Load in any observer parameters if specified
  if([FearlusArguments observerFileHasBeenSpecified]) {
    [ObjectLoader load: self fromFileNamed: [FearlusArguments getObserverFile]];
  }
  // Build the parameter object
  parameter = [Parameter createBegin: self];
  if([MiscFunc fileExists: [FearlusArguments getModelFile]]) {
				// Test to see if a model file exists before
                                // loading it.
    if([FearlusArguments debugObserver]) {
      printf("o: Loading model parameters from file %s\n",
	     [FearlusArguments getModelFile]);
    }
    [parameter loadFromFileNamed: [FearlusArguments getModelFile]];
    dumpHistory = [parameter dumpHistory];
    writeHistoryFrequency = [parameter maxHistoryLength];
  }
  else {
    fprintf(stderr, "Model parameter file %s not found -- using default values"
	    " for parameters\n", [FearlusArguments getModelFile]);
    [parameter setDumpHistory: dumpHistory];
    [parameter setMaxHistoryLength: writeHistoryFrequency];
  }
  if([FearlusArguments seedHasBeenSpecified]) {
    [parameter setSpecifiedSeed: seed];
  }
  parameter = [parameter createEnd];


  CREATE_PROBE_DISPLAY(parameter);
  //  CREATE_PROBE_DISPLAY(modelSwarm);
  CREATE_PROBE_DISPLAY(self);

  [self buildGUIObjects];

  [controlPanel setStateStopped];
  if([controlPanel getState] == ControlStateQuit) {
    exit(0);
  }

  [parameter finalise];
  // Build the model swarm -- this should not happen until the parameter object
  // has been finalised.
  modelSwarm = [ModelSwarm createBegin: self];
  [modelSwarm setParameters: parameter];
  modelSwarm = [modelSwarm createEnd];

  [parameter setDumpHistory: dumpHistory];
  [parameter setMaxHistoryLength: writeHistoryFrequency];
  [modelSwarm buildObjects];

  colorMap = [DBColormap create: self];

  [colorMap setColor: 0 ToName: "Gold"];
  for(i = 1; i < 40; i++) {
    r = g = b = ((double)i * 6.0) / 256.0;
    if(![colorMap setColor: i ToRed: r Green: g Blue: b]) {
				// Colours 1-39 are stepped grey levels, for
				// drawing graphs
      break;
    }
  }
  /* There now follows an attempt to create a colourmap of consecutively
     different colours -- each one trying to be different from the next one
     The results are indeed all different -- this much has been confirmed,
     but eyeballing them is not overly satisfactory... The first 30 or so
     colours are reasonably distinguishable from one another */
  while((i >= 40) && (i < 256)) {
    /* The tones array contains an array of tones alternating between
       light, dark, and mid tones, as fractions. The fractions are
       worked out from integer_tones = { 255, 95, 175, 215, 135, 235,
       115, 195 } by adding 1 and dividing by 256. */
    double tones[8] = { 1.0, 0.375, 0.6875, 0.84375,
			0.53125, 0.921875, 0.453125, 0.765625 };
    int j = i - 40;
    int tone = (j < 6) ? 0 : (j < 18) ? 1 : (j < 36) ? 2 : ((j / 36) + 2);
    double tone0 = 0.0, tone1 = 0.0, tone2 = 0.0;
    int step = (j < 6) ? 0 : (j < 18) ? (j - 6) / 6 : (j < 36) ? (j - 18) / 6
      : (j % 36) / 6;
    int s = j % 6;

    switch(step) {
    case 0:
      if(s < 3) {
	tone0 = tones[tone];
	tone1 = tone2 = 0.0;
      }
      else {
	tone0 = tone1 = tones[tone];
	tone2 = 0.0;
      }
      break;
    case 1:
      tone0 = tones[tone];
      tone1 = tones[tone - 1];
      tone2 = 0.0;
      break;
    case 2:
      tone0 = tones[tone];
      tone1 = tones[tone - 1];
      tone2 = tones[tone - 2];
      break;
    case 3:
      tone0 = tones[tone];
      tone1 = tones[tone - 3];
      tone2 = 0.0;
      break;
    case 4:
      tone0 = tones[tone];
      tone1 = tones[tone - 1];
      tone2 = tones[tone - 3];
      break;
    case 5:
      tone0 = tones[tone];
      tone1 = tones[tone - 2];
      tone2 = tones[tone - 3];
    }
    switch(s) {
    case 0:
      r = tone0; g = tone1; b = tone2;
      break;
    case 1:
      r = tone2; g = tone0; b = tone1;
      break;
    case 2:
      r = tone1; g = tone2; b = tone0;
      break;
    case 3:
      r = tone1; g = tone0; b = tone2;
      break;
    case 4:
      r = tone2; g = tone1; b = tone0;
      break;
    case 5:
      r = tone0; g = tone2; b = tone1;
      break;
    }
    if(![colorMap setColor: i ToRed: r Green: g Blue: b]) {
      break;
    }
    //    printf("%d %d %d %d %g %g %g\n", i, j, tone, step, r, g, b);
    //	   (int)((r * 256.0) - ((r == 0.0) ? 0.0 : 1.0)), 
    //	   (int)((g * 256.0) - ((g == 0.0) ? 0.0 : 1.0)),
    //	   (int)((b * 256.0) - ((b == 0.0) ? 0.0 : 1.0)));
    i++;
  }

  /* Now we want to create an array of pointers to char * so we can set the
     land use graph to have the same colours as in the raster */

  nColours = i;
  if(nColours < 40) {
    startColour = 0;
  }
  else {
    startColour = 40;
    nColours -= 40;
  }
  colours = (char **)malloc(nColours * sizeof(char *));
  if(colours == NULL) {
    [OutOfMemory raiseEvent];
  }
  colstr = [String create: scratchZone setC: "#FFFFFF"];
  for(i = startColour; i < startColour + nColours; i++) {
    [colorMap getColor: i inString: colstr];
    colours[i - startColour] = (char *)malloc(([colstr getCount] + 1)
					      * sizeof(char));
    if(!colours[i - startColour]) {
      [OutOfMemory raiseEvent];
    }
    strcpy(colours[i - startColour], [colstr getC]);
  }

  /* If we're displaying the subpop raster, then set the subpopulation
     colours.  The idea is you will enter your subpopulations in the
     directory as having colours starting at 0. A separate subpop
     colour string array is needed to set the colours of the subpop
     time-series graphs to match up to the colours on the subpop
     raster. This is because a given directory might contain a number
     of subpops and a number of model files. Each model file has a
     two-subpop contest, but you want consistent representation of the
     subpops across all modelfiles. We need to ask the subpops in this
     run what colours they are, and set the subpop colour array
     accordingly. This will only work properly if the subpops in the
     directory have their colours numbered from 0... */

  subpopcolours = (char **)malloc([parameter nSubPopulations]
				  * sizeof(char *));
  if(subpopcolours == NULL) {
    [OutOfMemory raiseEvent];
  }
  /* To make the subpops stand out from the land uses, if we have enough
     colours, start the subpops from the end of the colours used for the land
     uses, if there are enough colours available. */
  subpopStartColour 
    = ([parameter nUses] + [parameter nSubPopulations] > nColours) 
    ? startColour
    : startColour + [parameter nUses];
  for(i = 0; i < [parameter nSubPopulations]; i++) {
    SubPopulation *sp = [parameter subPopulation: (unsigned)i];
    subpopcolours[i] = colours[[sp getColour] 
			      + (subpopStartColour - startColour)];
    [sp setColour: [sp getColour] + subpopStartColour];
  }
  if([FearlusArguments debugObserver]) {
    printf("o: Allocated %d colours\n", nColours + startColour - 1);
  }
  [[modelSwarm getEnvironment] setMaxColour: nColours + startColour - 1];
  [[modelSwarm getLandAllocator] setMinColour: startColour
				 max: nColours + startColour - 1];
  
  // Create the colour map for the strategy display
  legend = [StrategyColourKey create: self];
  strategyCMap = [legend getColourMap];
  if(showStrategyColourKey) [legend display];
  [parameter setStrategyColourKey: legend];

  // [modelSwarm initialiseObjects];
  // Initialisation of objects is now done through the schedule
				// Now everything that needs to know how many
                                // colours there are knows how many colours
				// there are, they can get creating.


  // Allocate space for the lpBitStringRaster
  for(i = 1, j = 1; i < [parameter parcelBitStringSize]; i *= 10, j++) {}
  lpBitStringTitle = (char *)malloc(strlen(LPBITSTRINGRASTERTITLE) + j + 2);
  sprintf(lpBitStringTitle, "%s%d", LPBITSTRINGRASTERTITLE, 0);
  [[[modelSwarm getEnvironment] getLandParcels]
    forEach: M(setBitToDisplay:) : 0];
  lpBitStringBit = 0;

  // Decide the initial zoom factor, on the basis of the larger of the
  // environment's dimensions. Aim for a raster of about 250 x 250 pixels.

  maxEnvDim = [[modelSwarm getEnvironment] getSizeX];
  if([[modelSwarm getEnvironment] getSizeY] > maxEnvDim) {
    maxEnvDim = [[modelSwarm getEnvironment] getSizeY];
  }
  initZoom = 250 / maxEnvDim;

  // Create a list to store the widgets of all the graphs created

  widgetList = [List create: self];
  
  // Create the raster for the land use display

  if(showUseRaster) {
    envRaster = [ZoomRaster create: self];
    [envRaster setColormap: [colorMap getColormap]];
    [envRaster setZoomFactor: initZoom * zoomSize];
    [envRaster setWidth: [[modelSwarm getEnvironment] getSizeX]
	       Height: [[modelSwarm getEnvironment] getSizeY]];
    [envRaster setWindowTitle: "Land Parcels by Land Use"];
    [envRaster pack];		// This apparently means draw it
  }

  // Create the raster for showing the LP bitstrings

  if(showLPBitStringRaster) {
    lpBitStringRaster = [ZoomRaster create: self];
    [lpBitStringRaster setColormap: [colorMap getColormap]];
    [lpBitStringRaster setZoomFactor: initZoom * zoomSize];
    [lpBitStringRaster setWidth: [[modelSwarm getEnvironment] getSizeX]
		       Height: [[modelSwarm getEnvironment] getSizeY]];
    [lpBitStringRaster setWindowTitle: lpBitStringTitle];
    [lpBitStringRaster pack];
  }

  // Create the raster for the land use change display

  if(showUseChangeRaster) {
    landUseChangeRaster = [ZoomRaster create: self];
    [landUseChangeRaster setColormap: [colorMap getColormap]];
    [landUseChangeRaster setZoomFactor: initZoom * zoomSize];
    [landUseChangeRaster setWidth: [[modelSwarm getEnvironment] getSizeX]
			 Height: [[modelSwarm getEnvironment] getSizeY]];
    [landUseChangeRaster setWindowTitle:
			   "Land Parcels by Land Use Change"];
    [landUseChangeRaster pack];
  }

  // Create the raster for the land manager display

  if(showManagerRaster) {
    landMgrRaster = [ZoomRaster create: self];
    [landMgrRaster setColormap: [colorMap getColormap]];
    [landMgrRaster setZoomFactor: initZoom * zoomSize];
    [landMgrRaster setWidth: [[modelSwarm getEnvironment] getSizeX]
		   Height: [[modelSwarm getEnvironment] getSizeY]];
    [landMgrRaster setWindowTitle: "Land Parcels by Land Manager"];
    [landMgrRaster pack];
  }

  // Create the raster for the subpopulation display

  if(showSubPopRaster) {
    subPopRaster = [ZoomRaster create: self];
    [subPopRaster setColormap: [colorMap getColormap]];
    [subPopRaster setZoomFactor: initZoom * zoomSize];
    [subPopRaster setWidth: [[modelSwarm getEnvironment] getSizeX]
		  Height: [[modelSwarm getEnvironment] getSizeY]];
    [subPopRaster setWindowTitle: "Land Parcels by Sub-Population"];
    [subPopRaster pack];
  }

  // Create the raster for the land manager change display

  if(showManagerChangeRaster) {
    landMgrChangeRaster = [ZoomRaster create: self];
    [landMgrChangeRaster setColormap: [colorMap getColormap]];
    [landMgrChangeRaster setZoomFactor: initZoom * zoomSize];
    [landMgrChangeRaster setWidth: [[modelSwarm getEnvironment] getSizeX]
			 Height: [[modelSwarm getEnvironment] getSizeY]];
    [landMgrChangeRaster 
      setWindowTitle: "Land Parcels by Land Manager Change"];
    [landMgrChangeRaster pack];
  }

  // Create the raster for the strategy display

  if(showStrategyRaster) {
    strategyRaster = [ZoomRaster create: self];
    [strategyRaster setColormap: strategyCMap];
    [strategyRaster setZoomFactor: initZoom * zoomSize];
    [strategyRaster setWidth: [[modelSwarm getEnvironment] getSizeX]
		    Height: [[modelSwarm getEnvironment] getSizeY]];
    [strategyRaster setWindowTitle: "Parcels by Strategy used to select Use"];
    [strategyRaster pack];
  }

  // Create the land use display

  if(showUseRaster) {
    landParcelDisplay = [Object2dDisplay createBegin: self];
    [landParcelDisplay setDisplayWidget: envRaster];
    [landParcelDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [landParcelDisplay setObjectCollection:
			 [[modelSwarm getEnvironment] getLandParcels]];
    [landParcelDisplay setDisplayMessage: M(drawSelfOn:)];
    landParcelDisplay = [landParcelDisplay createEnd];

    if(showLPBitStringDiffs) {
      parcelBoundaryDisplay = [Object2dDisplay createBegin: self];
      [parcelBoundaryDisplay setDisplayWidget: envRaster];
      [parcelBoundaryDisplay setDiscrete2dToDisplay:
			       [modelSwarm getEnvironment]];
      [parcelBoundaryDisplay setObjectCollection:
			       [[modelSwarm getEnvironment] getLandParcels]];
      [parcelBoundaryDisplay setDisplayMessage: M(drawMatchLinesOn:)];
      parcelBoundaryDisplay = [parcelBoundaryDisplay createEnd];
    }

    parcelNbrDisplay = [Object2dDisplay createBegin: self];
    [parcelNbrDisplay setDisplayWidget: envRaster];
    [parcelNbrDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [parcelNbrDisplay setObjectCollection:
			[[modelSwarm getEnvironment] getLandParcels]];
    [parcelNbrDisplay setDisplayMessage: M(drawNbrLinesOn:)];
    parcelNbrDisplay = [parcelNbrDisplay createEnd];
  }

  // Create the LP bit string display

  if(showLPBitStringRaster) {
    lpBitStringDisplay = [Object2dDisplay createBegin: self];
    [lpBitStringDisplay setDisplayWidget: lpBitStringRaster];
    [lpBitStringDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [lpBitStringDisplay setObjectCollection:
			  [[modelSwarm getEnvironment] getLandParcels]];
    [lpBitStringDisplay setDisplayMessage: M(drawBitStringOn:)];
    lpBitStringDisplay = [lpBitStringDisplay createEnd];
  }

  // Create the land use change display

  if(showUseChangeRaster) {
    landUseChangeDisplay = [Object2dDisplay createBegin: self];
    [landUseChangeDisplay setDisplayWidget: landUseChangeRaster];
    [landUseChangeDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [landUseChangeDisplay setObjectCollection:
			    [[modelSwarm getEnvironment] getLandParcels]];
    [landUseChangeDisplay setDisplayMessage: M(drawChangeLURankOn:)];
    landUseChangeDisplay = [landUseChangeDisplay createEnd];
  }

  // Create the land manager display

  if(showManagerRaster) {
    landMgrDisplay = [Object2dDisplay createBegin: self];
    [landMgrDisplay setDisplayWidget: landMgrRaster];
    [landMgrDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [landMgrDisplay setObjectCollection:
		      [[modelSwarm getEnvironment] getLandParcels]];
    [landMgrDisplay setDisplayMessage: M(drawMgrOn:)];
    landMgrDisplay = [landMgrDisplay createEnd];
  }

  // Create the subpopulation display

  if(showSubPopRaster) {
    subPopDisplay = [Object2dDisplay createBegin: self];
    [subPopDisplay setDisplayWidget: subPopRaster];
    [subPopDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [subPopDisplay setObjectCollection:
		     [[modelSwarm getEnvironment] getLandParcels]];
    [subPopDisplay setDisplayMessage: M(drawSubPopOn:)];
    subPopDisplay = [subPopDisplay createEnd];
  }

  // Create the land manager change display

  if(showManagerChangeRaster) {
    landMgrChangeDisplay = [Object2dDisplay createBegin: self];
    [landMgrChangeDisplay setDisplayWidget: landMgrChangeRaster];
    [landMgrChangeDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [landMgrChangeDisplay setObjectCollection:
			    [[modelSwarm getEnvironment] getLandParcels]];
    [landMgrChangeDisplay setDisplayMessage: M(drawChangeLMgrRankOn:)];
    landMgrChangeDisplay = [landMgrChangeDisplay createEnd];
  }

  // Create the strategy display

  if(showStrategyRaster) {
    strategyDisplay = [Object2dDisplay createBegin: self];
    [strategyDisplay setDisplayWidget: strategyRaster];
    [strategyDisplay setDiscrete2dToDisplay: [modelSwarm getEnvironment]];
    [strategyDisplay setObjectCollection:
		       [[modelSwarm getEnvironment] getLandParcels]];
    [strategyDisplay setDisplayMessage: M(drawStrategyOn:)];
    strategyDisplay = [strategyDisplay createEnd];
  }

  // Enable right button click on each of the rasters

  if(showUseRaster) {
    [envRaster setButton: ButtonMiddle Client: landParcelDisplay
	       Message: M(makeProbeAtX:Y:)];
    [envRaster setButton: ButtonLeft Client: envRaster
	       Message: M(increaseZoom)];
    [envRaster setButton: ButtonRight Client: envRaster
	       Message: M(decreaseZoom)];
  }
  if(showUseChangeRaster) {
    [landUseChangeRaster setButton: ButtonRight Client: landUseChangeDisplay
			 Message: M(makeProbeAtX:Y:)];
  }
  if(showManagerRaster) {
    [landMgrRaster setButton: ButtonRight Client: landMgrDisplay
		   Message: M(makeProbeAtX:Y:)];
  }
  if(showSubPopRaster) {
    [subPopRaster setButton: ButtonMiddle Client: subPopDisplay
		  Message: M(makeProbeAtX:Y:)];
    [subPopRaster setButton: ButtonLeft Client: subPopRaster
		  Message: M(increaseZoom)];
    [subPopRaster setButton: ButtonRight Client: subPopRaster
		  Message: M(decreaseZoom)];
  }
  if(showManagerChangeRaster) {
    [landMgrChangeRaster setButton: ButtonRight Client: landMgrChangeDisplay
			 Message: M(makeProbeAtX:Y:)];
  }
  if(showStrategyRaster) {
    [strategyRaster setButton: ButtonRight Client: strategyDisplay
		    Message: M(makeProbeAtX:Y:)];
  }

  // Enable left button click on some of the rasters

  // On the land manager raster, enable the user to display the physical and
  // social neighbourhood of a particular land manager
  if(showManagerRaster) {
    [landMgrRaster setButton: ButtonLeft Client: self
		   Message: M(toggleMgrNeighbourhoodOfX:Y:)];
  }

  // On the land parcel raster, enable the user to display the physical
  // neighbourhood of a particular cell
  if(showUseRaster) {
    [envRaster setButton: ButtonLeft Client: self
	       Message: M(togglePhysNeighbourhoodOfX:Y:)];
  }
  
  // On the land parcel bit string raster, enable the user to cycle the
  // bit displayed
  if(showLPBitStringRaster) {
    [lpBitStringRaster setButton: ButtonLeft Client: self
		       Message: M(decLandParcelBitDisplayed)];
    [lpBitStringRaster setButton: ButtonRight Client: self
		       Message: M(incLandParcelBitDisplayed)];
  }

  // Create the land manager land parcel graph

  if(showParcelsOwnedGraph) {
    parcelsOwnedGraph = [EZGraph createBegin: [self getZone]];
    [parcelsOwnedGraph setTitle:
			 "Land Parcels Owned by Land Managers"];
    [parcelsOwnedGraph setAxisLabelsX: "time" Y: "number of land parcels"];
    if([parameter nSubPopulations] > 1) {
      [parcelsOwnedGraph setColors: (const char * const *)subpopcolours
			 count: (unsigned)[parameter nSubPopulations]];
    }
    parcelsOwnedGraph = [parcelsOwnedGraph createEnd];
    [widgetList addLast: [parcelsOwnedGraph getGraph]];
  }

  // Create the land manager wealth graph

  if(showWealthGraph) {
    wealthGraph = [EZGraph createBegin: [self getZone]];
    [wealthGraph setTitle: "Wealth of Land Managers vs. time"];
    [wealthGraph setAxisLabelsX: "time" Y: "wealth"];
    if([parameter nSubPopulations] > 1) {
      [wealthGraph setColors: (const char * const *)subpopcolours
		   count: (unsigned)[parameter nSubPopulations]];
    }
    wealthGraph = [wealthGraph createEnd];
    [widgetList addLast: [wealthGraph getGraph]];
  }

  // Create the land manager age graph

  if(showAgeGraph) {
    ageGraph = [EZGraph createBegin: [self getZone]];
    [ageGraph setTitle: "Age of Land Managers vs. time"];
    [ageGraph setAxisLabelsX: "time" Y: "age"];
    if([parameter nSubPopulations] > 1) {
      [ageGraph setColors: (const char * const *)subpopcolours
		count: (unsigned)[parameter nSubPopulations]];
    }
    ageGraph = [ageGraph createEnd];
    [widgetList addLast: [ageGraph getGraph]];
  }

  // Create the yield graph

  if(showYieldGraph) {
    yieldGraph = [EZGraph createBegin: [self getZone]];
    [yieldGraph setTitle: "Yield of Land Parcels vs. time"];
    [yieldGraph setAxisLabelsX: "time" Y: "yield"];
    yieldGraph = [yieldGraph createEnd];
    [widgetList addLast: [yieldGraph getGraph]];
  }

  // Create the population graph

  if(showPopulationGraph) {
    landMgrPopGraph = [EZGraph createBegin: [self getZone]];
    [landMgrPopGraph setTitle: "Population of Land Managers vs. time"];
    [landMgrPopGraph setAxisLabelsX: "time" Y: "population"];
    [landMgrPopGraph setColors: (const char * const *)subpopcolours
		     count: (unsigned)[parameter nSubPopulations]];
    landMgrPopGraph = [landMgrPopGraph createEnd];
    [widgetList addLast: [landMgrPopGraph getGraph]];
  }

  // Create the land use graph

  if(showUseGraph) {
    landUseGraph = [EZGraph createBegin: [self getZone]];
    [landUseGraph setTitle: "Number of Land Parcels for each Land Use"];
    [landUseGraph setAxisLabelsX: "time" Y: "parcels"];
    [landUseGraph setColors: (const char * const *)colours
		  count: (unsigned)nColours];
    landUseGraph = [landUseGraph createEnd];
    [widgetList addLast: [landUseGraph getGraph]];
  }

  // Create the arrays for the various distribution graphs

  if(showNWeightDist) {
    nbrwgtDists = [List create: self];
  }
  if(showThresholdDist) {
    happyThreshDists = [List create: self];
  }
  if(showImitProbDist) {
    imitProbDists = [List create: self];
  }
  // We'll be adding the sequences later (once the land uses have been created)

  // Don't forget to add any new graphical display to the schedule in
  // buildActions

  return(self);
}

/*

toggleMgrNeighbourhoodOfX:Y:

Call the LandParcel at X, Y and tell it to display/undisplay the physical and
social neighbourhoods of the land manager owning the land parcel at X, Y.

*/

-toggleMgrNeighbourhoodOfX: (int)x Y: (int)y {
  [(LandParcel *)[[modelSwarm getEnvironment] 
		   getObjectAtX: x Y: y] toggleMgrNbrDisplay];
  return self;
}

/*

inc/decLandParcelBitDisplayed

Change the bit displayed by the bitstring raster

*/

-incLandParcelBitDisplayed {
  lpBitStringBit = [MiscFunc Mod: (lpBitStringBit + 1)
			     inRange0To: [parameter parcelBitStringSize]];
  sprintf(lpBitStringTitle, "%s%d", LPBITSTRINGRASTERTITLE, lpBitStringBit);
  [[[modelSwarm getEnvironment] getLandParcels] forEach: M(incBitToDisplay)];
  [lpBitStringRaster setWindowTitle: lpBitStringTitle];
  [lpBitStringDisplay display];
  [lpBitStringRaster drawSelf];
  return(self);
}

-decLandParcelBitDisplayed {
  lpBitStringBit = [MiscFunc Mod: (lpBitStringBit - 1)
			     inRange0To: [parameter parcelBitStringSize]];
  sprintf(lpBitStringTitle, "%s%d", LPBITSTRINGRASTERTITLE, lpBitStringBit);
  [[[modelSwarm getEnvironment] getLandParcels] forEach: M(decBitToDisplay)];
  [lpBitStringRaster setWindowTitle: lpBitStringTitle];
  [lpBitStringDisplay display];
  [lpBitStringRaster drawSelf];
  return(self);
}

/*

togglePhysNeighbourhoodOfX:Y:

Call the LandParcel at X, Y and tell it to display/undisplay its physical 
neighbourhood.

*/

-togglePhysNeighbourhoodOfX: (int)x Y: (int)y {
  [(LandParcel *)[[modelSwarm getEnvironment]
		   getObjectAtX: x Y: y] togglePhysNbrDisplay];
  return self;
}

/*

populateGraphs

Do any actions for the graphs that require things to have been created.

*/

-(void)populateGraphs {
  int i;
  id landUses;

  // Populate the land use graph
  if(showUseGraph) {
    landUses = [[modelSwarm getEnvironment] getLandUses];
    for(i = 0; i < [landUses getCount]; i++) {
      [landUseGraph createSequence: [[landUses atOffset: i] getLabel]
		    withFeedFrom:	[landUses atOffset: i]
		    andSelector: M(getNLandParcels)];
    }
  }
  // Populate the parcels owned graph
  if([parameter nSubPopulations] == 1) {
    if(showParcelsOwnedGraph) {
      [parcelsOwnedGraph createAverageSequence: "average"
			 withFeedFrom:
			   [[modelSwarm getLandAllocator] getLandManagers]
			 andSelector: M(getNumberOfLandParcelsOwned)];
      [parcelsOwnedGraph createMaxSequence: "maximum"
			 withFeedFrom:
			   [[modelSwarm getLandAllocator] getLandManagers]
			 andSelector: M(getNumberOfLandParcelsOwned)];
      [parcelsOwnedGraph createMinSequence: "minimum"
			 withFeedFrom:
			   [[modelSwarm getLandAllocator] getLandManagers]
			 andSelector: M(getNumberOfLandParcelsOwned)];
    }
    // Populate the wealth graph
    if(showWealthGraph) {
      [wealthGraph createAverageSequence: "average"
		   withFeedFrom: 
		     [[modelSwarm getLandAllocator] getLandManagers]
		   andSelector: M(getBarnAsDouble)];
      [wealthGraph createMaxSequence: "maximum"
		   withFeedFrom: 
		     [[modelSwarm getLandAllocator] getLandManagers]
		   andSelector: M(getBarnAsDouble)];
      [wealthGraph createMinSequence: "minimum"
		   withFeedFrom:
		     [[modelSwarm getLandAllocator] getLandManagers]
		   andSelector: M(getBarnAsDouble)];
    }
    // Populate the age graph
    if(showAgeGraph) {
      [ageGraph createAverageSequence: "average"
		withFeedFrom: [[modelSwarm getLandAllocator] getLandManagers]
		andSelector: M(getAgeAsInt)];
      [ageGraph createMaxSequence: "maximum"
		withFeedFrom: [[modelSwarm getLandAllocator] getLandManagers]
		andSelector: M(getAgeAsInt)];
      [ageGraph createMinSequence: "minimum"
		withFeedFrom: [[modelSwarm getLandAllocator] getLandManagers]
		andSelector: M(getAgeAsInt)];
    }
  }
  else {
    for(i = 0; i < [parameter nSubPopulations]; i++) {
      SubPopulation *sp;

      sp = [parameter subPopulation: i];

      if(showParcelsOwnedGraph) {
	[parcelsOwnedGraph createTotalSequence: [sp getLabel]
			   withFeedFrom: [sp getLandManagers]
			   andSelector: M(getNumberOfLandParcelsOwned)];
      }
      if(showWealthGraph) {
	[wealthGraph createAverageSequence: [sp getLabel]
		     withFeedFrom: [sp getLandManagers]
		     andSelector: M(getBarnAsDouble)];
      }
      if(showAgeGraph) {
	[ageGraph createAverageSequence: [sp getLabel]
		  withFeedFrom: [sp getLandManagers]
		  andSelector: M(getAgeAsInt)];
      }
    }
  }
  // Populate the yield graph
  if(showYieldGraph) {
#ifdef SOLVED_AVERAGE_BUG
    [yieldGraph createAverageSequence: "average"
		withFeedFrom: [[modelSwarm getEnvironment] getLandParcels]
		andSelector: M(getIntYield)];
#else
    [yieldGraph createTotalSequence: "total"
		withFeedFrom: [[modelSwarm getEnvironment] getLandParcels]
		andSelector: M(getIntYield)];
#endif
    [yieldGraph createMaxSequence: "maximum"
		withFeedFrom: [[modelSwarm getEnvironment] getLandParcels]
		andSelector: M(getIntYield)];
    [yieldGraph createMinSequence: "minimum"
		withFeedFrom: [[modelSwarm getEnvironment] getLandParcels]
		andSelector: M(getIntYield)];
  }

  // Populate the population graph

  if(showPopulationGraph) {
    for(i = 0; i < [parameter nSubPopulations]; i++) {
      [landMgrPopGraph createSequence: [[parameter subPopulation: i] getLabel]
		       withFeedFrom: [parameter subPopulation: i]
		       andSelector: M(getManagerCount)];
    }
  }

  // Create the neighbourhood weighting histogram(s)

  if(showNWeightDist) {
    id <EZBin> nbrwgtDist;

    for(i = 0; i < [parameter nSubPopulations]; i++) {
      double lower, upper;

      if(strcmp([[parameter subPopulation: i] getNeighbourWeightDist],
		"uniform") == 0) {
	lower = [[parameter subPopulation: i] getNeighbourWeightMin];
	upper = [[parameter subPopulation: i] getNeighbourWeightMax];
      }
      else {
	double mean, var;
	mean = [[parameter subPopulation: i] getNeighbourWeightMean];
	var = [[parameter subPopulation: i] getNeighbourWeightVar];
	lower = mean - (NDIST_SD_MULTIPLIER * sqrt(var));
	upper = mean + (NDIST_SD_MULTIPLIER * sqrt(var));
      }
      if(lower == upper) {
	fprintf(stderr, "Not creating a neighbourhood weighting distribution "
		"graph for %s because lower bound %g = upper bound %g\n",
		[[parameter subPopulation: i] getLabel], lower, upper);
	continue;
      }
      nbrwgtDist = [EZBin createBegin: self];
      [nbrwgtDist setTitle: [[parameter subPopulation: i] getLabel]];
      [nbrwgtDist setAxisLabelsX: "neighbourhood weighting" Y: "#managers"];
      [nbrwgtDist setMonoColorBars: YES];
      [nbrwgtDist setBinCount: nHistogramBins];
      [nbrwgtDist setLowerBound: lower];
      [nbrwgtDist setUpperBound: upper];
      [nbrwgtDist setCollection:
		    [[parameter subPopulation: i] getLandManagers]];
      [nbrwgtDist setProbedSelector: M(getNeighbourWeight)];
      nbrwgtDist = [nbrwgtDist createEnd];
      [widgetList addLast: [nbrwgtDist getHistogram]];
      [nbrwgtDists addLast: nbrwgtDist];
    }
  }

  // Create the happiness threshold histogram(s)

  if(showThresholdDist) {
    id <EZBin> happyThreshDist;

    for(i = 0; i < [parameter nSubPopulations]; i++) {
      double lower, upper;

      if(strcmp([[parameter subPopulation: i] getHabitThresholdDist],
		"uniform") == 0) {
	lower = [[parameter subPopulation: i] getHabitThresholdMin];
	upper = [[parameter subPopulation: i] getHabitThresholdMax];
				// N.B. "habit" here refers to aspiration
      }
      else {
	double mean, var;
	mean = [[parameter subPopulation: i] getHabitThresholdMean];
	var = [[parameter subPopulation: i] getHabitThresholdVar];
	lower = mean - (NDIST_SD_MULTIPLIER * sqrt(var));
	upper = mean + (NDIST_SD_MULTIPLIER * sqrt(var));
      }
      if(lower == upper) {
	fprintf(stderr, "Not creating a happiness threshold distribution "
		"graph for %s because lower bound %g = upper bound %g\n",
		[[parameter subPopulation: i] getLabel], lower, upper);
	continue;
      }
      happyThreshDist = [EZBin createBegin: self];
      [happyThreshDist setTitle: [[parameter subPopulation: i] getLabel]];
      [happyThreshDist setAxisLabelsX: "happiness threshold"
		       Y: "#managers"];
      [happyThreshDist setMonoColorBars: YES];
      [happyThreshDist setBinCount: nHistogramBins];
      [happyThreshDist setLowerBound: lower];
      [happyThreshDist setUpperBound: upper];
      [happyThreshDist setCollection:
			 [[parameter subPopulation: i] getLandManagers]];
      [happyThreshDist setProbedSelector: M(getHabitThreshold)];
      happyThreshDist = [happyThreshDist createEnd];
      [widgetList addLast: [happyThreshDist getHistogram]];
      [happyThreshDists addLast: happyThreshDist];
				// N.B. "happy" and "habit" refer to aspiration
    }
  }

  // Create the imitative probability histogram(s)

  if(showImitProbDist) {
    id <EZBin> imitProbDist;

    for(i = 0; i < [parameter nSubPopulations]; i++) {
      double lower, upper;

      if(strcmp([[parameter subPopulation: i] getImitateProbDist],
		"uniform") == 0) {
	lower = [[parameter subPopulation: i] getImitateProbMin];
	upper = [[parameter subPopulation: i] getImitateProbMax];
      }
      else {
	double mean, var;
	mean = [[parameter subPopulation: i] getImitateProbMean];
	var = [[parameter subPopulation: i] getImitateProbVar];
	lower = mean - (NDIST_SD_MULTIPLIER * sqrt(var));
	upper = mean + (NDIST_SD_MULTIPLIER * sqrt(var));
      }
      if(lower == upper) {
	fprintf(stderr, "Not creating an imitate probability distribution "
		"graph for %s because lower bound %g = upper bound %g\n",
		[[parameter subPopulation: i] getLabel], lower, upper);
	continue;
      }
      imitProbDist = [EZBin createBegin: self];
      [imitProbDist setTitle: [[parameter subPopulation: i] getLabel]];
      [imitProbDist setAxisLabelsX: "imitate probability"
		       Y: "#managers"];
      [imitProbDist setMonoColorBars: YES];
      [imitProbDist setBinCount: nHistogramBins];
      [imitProbDist setLowerBound: lower];
      [imitProbDist setUpperBound: upper];
      [imitProbDist setCollection:
		    [[parameter subPopulation: i] getLandManagers]];
      [imitProbDist setProbedSelector: M(getImitateProb)];
      imitProbDist = [imitProbDist createEnd];
      [widgetList addLast: [imitProbDist getHistogram]];
      [imitProbDists addLast: imitProbDist];
    }
  }
}

-resizeGraphs {
  id <Index> inx;

  for(inx = [widgetList begin: self], [inx next];
      [inx getLoc] == Member;
      [inx next]) {
    id <Widget> w = (id <Widget>)[inx get];
    [w setWidth: [w getWidth] * zoomSize Height: [w getHeight] * zoomSize];
  }
  [inx drop];
  [widgetList drop];
  [displayActions remove: resizeGraphsAction];
  return self;
}

/*

buildActions

Override the Swarm buildActions method to introduce the scheduling required
for this class.

*/

-buildActions {
  [super buildActions];
  [modelSwarm buildActions];
  displayActions = [ActionGroup create: self];

  [displayActions createActionTo: [modelSwarm getEnvironment]
		  message: M(rankLandParcels)];
  if(showUseRaster) {
    [displayActions
      createActionTo: landParcelDisplay 	message: M(display)];
    if(showLPBitStringDiffs) {
      [displayActions
	createActionTo: parcelBoundaryDisplay 	message: M(display)];
    }
    [displayActions
      createActionTo: parcelNbrDisplay		message: M(display)];
    [displayActions
      createActionTo: envRaster 		message: M(drawSelf)];
  }
  if(showLPBitStringRaster) {
    [displayActions
      createActionTo: lpBitStringDisplay	message: M(display)];
    [displayActions
      createActionTo: lpBitStringRaster		message: M(drawSelf)];
  }
  if(showUseChangeRaster) {
    [displayActions
      createActionTo: landUseChangeDisplay 	message: M(display)];
    [displayActions
      createActionTo: landUseChangeRaster 	message: M(drawSelf)];
  }
  if(showManagerChangeRaster) {
    [displayActions
      createActionTo: landMgrChangeDisplay 	message: M(display)];
    [displayActions
      createActionTo: landMgrChangeRaster 	message: M(drawSelf)];
  }
  if(showManagerRaster) {
    [displayActions
      createActionTo: landMgrDisplay	 	message: M(display)];
    [displayActions
      createActionTo: landMgrRaster 		message: M(drawSelf)];
  }
  if(showSubPopRaster) {
    [displayActions
      createActionTo: subPopDisplay             message: M(display)];
    [displayActions
      createActionTo: subPopRaster              message: M(drawSelf)];
  }
  if(showStrategyRaster) {
    [displayActions
      createActionTo: strategyDisplay		message: M(display)];
    [displayActions
      createActionTo: strategyRaster	 	message: M(drawSelf)];
  }
  if(showParcelsOwnedGraph) {
    [displayActions
      createActionTo: parcelsOwnedGraph		message: M(step)];
  }
  if(showWealthGraph) {
    [displayActions
      createActionTo: wealthGraph 		message: M(step)];
  }
  if(showAgeGraph) {
    [displayActions
      createActionTo: ageGraph	 		message: M(step)];
  }
  if(showYieldGraph) {
    [displayActions
      createActionTo: yieldGraph 		message: M(step)];
  }
  if(showPopulationGraph) {
    [displayActions
      createActionTo: landMgrPopGraph	 	message: M(step)];
  }
  if(showUseBar) {
    [displayActions
      createActionTo: landUseBar		message: M(step)];
  }
  if(showUseGraph) {
    [displayActions
      createActionTo: landUseGraph		message: M(step)];
  }
  if(showNWeightDist) {
    [displayActions
      createActionForEach: nbrwgtDists		message: M(reset)];
    [displayActions
      createActionForEach: nbrwgtDists		message: M(update)];
    [displayActions
      createActionForEach: nbrwgtDists		message: M(output)];
  }
  if(showThresholdDist) {
    [displayActions
      createActionForEach: happyThreshDists	message: M(reset)];
    [displayActions
      createActionForEach: happyThreshDists	message: M(update)];
    [displayActions
      createActionForEach: happyThreshDists	message: M(output)];
  }
  if(showImitProbDist) {
    [displayActions
      createActionForEach: imitProbDists	message: M(reset)];
    [displayActions
      createActionForEach: imitProbDists	message: M(update)];
    [displayActions
      createActionForEach: imitProbDists	message: M(output)];
  }
  [displayActions createActionTo: probeDisplayManager 	message: M(update)];
				// If we do more displays, they'll have to
				// be added in here
  [displayActions createActionTo: actionCache message: M(doTkEvents)];
  resizeGraphsAction = [displayActions createActionTo: self
				       message: M(resizeGraphs)];
  displaySchedule = [Schedule createBegin: self];
  [displaySchedule setRepeatInterval: displayFrequency];
  displaySchedule = [displaySchedule createEnd];
  [displaySchedule at: 0 createAction: displayActions];

  historySchedule = [Schedule createBegin: self];
  [historySchedule setRepeatInterval: writeHistoryFrequency];
  historySchedule = [historySchedule createEnd];
  [historySchedule at: 0 createActionTo: self
		   message: M(writeHistoryToFile:) : (id)historyFileName];


  initialSchedule = [Schedule createBegin: self];
  initialSchedule = [initialSchedule createEnd];
  [initialSchedule at: 0 createActionTo: self message: M(populateGraphs)];
  [initialSchedule at: 0 createActionTo: displaySchedule
		   message: M(activateIn:) : self];
  if([FearlusArguments useReporter]) {
    [initialSchedule at: 0 createActionTo: self message: M(buildReporter)];
  }
  if(saveDumpedHistory) {
    [initialSchedule at: writeHistoryFrequency - 1
		     createActionTo: historySchedule
		     message: M(activateIn:) : self];
  }
  if(![parameter infiniteTime]) {
    [initialSchedule at: [parameter maximumYear]
		     createActionTo: controlPanel
		     message: M(setStateStopped)];
  }


  return(self);
}

/*

activateIn

Override the Swarm activateIn method to activate the schedules

*/

-activateIn: (id)swarmContext {
  [super activateIn: swarmContext];
  [modelSwarm activateIn: self];	// model schedule must come first
  // (This means it will get done first, so the display always occurs after
  // the model cycle)
  [initialSchedule activateIn: self];
  return([self getSwarmActivity]);
}

/*

setSpecifiedSeed:

If the user specified a seed on the command line, then it should be recorded
in the history file. So it needs to be passed in from main.

*/

-(void)setSpecifiedSeed: (unsigned)value {
  seed = value;
}

/*

writeHistory

Call the modelSwarm to get it to write its history to a file that has been
created. This should be made available as a button so it can be called at any
time. If this method has a return type of void, it cannot be made available as
a button. (You get an abort when you click it.)

If the user has specified on the command line that they want a history file
created, then clicking the button will have no effect. A full history will be
written when they quit the application.

*/

-writeHistory {
  if([FearlusArguments historyFileHasBeenSpecified]) {
    fprintf(stderr, "History will be written when you quit\n");
  }
  else {
    [self writeHistoryToFile: historyFileName];
  }
  return self;
}

/*

writeHistoryToFile

The method that actually writes the history to a file. Empty file names will
not be written to. This method enforces a .txt suffix for XL. If the file 
exists, it will insert the minimum three digit number before the .txt suffix
required to make the filename unique.

*/

-(void)writeHistoryToFile: (const char *)fileName {
  char *realFileName;
  FILE *fp;

  realFileName = [MiscFunc getUsableFileNameFrom: fileName withSuffix: ".txt"];
  if(realFileName == NULL) {
    fprintf(stderr, "%s -- cannot write history because couldn't create "
	    "datafile\n", sel_get_name(_cmd));
    return;
  }
  // Now we actually write the file
  fp = fopen(realFileName, "w");
  if(!fp) {
    fprintf(stderr, "Cannot create datafile %s\n", realFileName);
    perror(realFileName);
  }
  else {
    printf("Saving data to file: %s\n", realFileName);
    [modelSwarm writeHistory: fp];
    fclose(fp);
  }
  free(realFileName);
}

/*

drop

Override the Swarm drop method to make sure all items created by this object
are dropped

*/

-(void)drop {
  [landMgrDisplay drop];
  [landParcelDisplay drop];
  [landMgrChangeDisplay drop];
  [landUseChangeDisplay drop];
  [strategyDisplay drop];
  //  [landUseBar drop];
  [legend drop];
  [landMgrPopGraph drop];
  [yieldGraph drop];
  [wealthGraph drop];
  [ageGraph drop];
  [parcelsOwnedGraph drop];
  [landMgrRaster drop];
  [landMgrChangeRaster drop];
  [landUseChangeRaster drop];
  [envRaster drop];
  [strategyRaster drop];
  [modelSwarm drop];
  free(historyFileName);
  [super drop];
}

@end
