/*
    FEARLUS model0-6-2: main.m
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute
    Copyright (C) 1999-2002  Macaulay Institute

    This file is part of FEARLUS model0-6-2, an agent-based model of land use
    change.

    FEARLUS model0-6-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


The main function. This is copied straight out of Heatbugs, and modified for
the FEARLUS objects.

*/

// Heatbugs application. Copyright (C) 1996-1998 Santa Fe Institute.
// This library is distributed without any warranty; without even the
// implied warranty of merchantability or fitness for a particular purpose.
// See file LICENSE for details and terms of copying.

#import <simtools.h>     // initSwarm () and swarmGUIMode
#import <simtoolsgui.h>  // GUISwarm
#import <random.h>
#import <objc/objc-api.h> // objc_trace
#import <objc/Protocol.h>
#import "ObserverSwarm.h"
#import "BatchSwarm.h"
#import "FearlusArguments.h"
#import "Strategy.h"
#import "Report.h"
#import "Clumper.h"
#import "ClassInfo.h"


// The main() function is the top-level place where everything starts.
// For a typical Swarm simulation, in main() you create a toplevel
// Swarm, let it build and activate, and set it to running.

int
main (int argc, const char **argv)
{
  id theTopLevelSwarm;
  unsigned seed;

  
  // Swarm initialization: all Swarm apps must call this first.
  initSwarmArguments (argc, argv, [FearlusArguments class]);

  [ClassInfo initialise: [Zone create: globalZone]];
  seed = [randomGenerator getInitialSeed];
  if([FearlusArguments seedHasBeenSpecified]) {
    seed = [FearlusArguments getSeedArg];
    printf("Using seed: %u\n", seed);
    [randomGenerator setStateFromSeed: seed];
  }
  else {
    printf("Using default Swarm seed: %u\n", seed);
  }

  // swarmGUIMode is set in initSwarmArguments(). It's set to be 0 if
  // you typed heatbugs -batchmode. Otherwise, it's set to 1.
  
  if (swarmGUIMode == 1) {
    // We've got graphics, so make a full ObserverSwarm to get GUI objects

    theTopLevelSwarm = [ObserverSwarm createBegin: globalZone];
    SET_WINDOW_GEOMETRY_RECORD_NAME (theTopLevelSwarm);
    theTopLevelSwarm = [theTopLevelSwarm createEnd];
  }
  else {
    // No graphics - make a batchmode swarm and run it.

    theTopLevelSwarm = [BatchSwarm create: globalZone];
  }

  if([FearlusArguments seedHasBeenSpecified]) {
    [theTopLevelSwarm setSpecifiedSeed: seed];
  }
  [theTopLevelSwarm buildObjects];
  [theTopLevelSwarm buildActions];
  [theTopLevelSwarm activateIn: nil];
  [theTopLevelSwarm go];
  if([FearlusArguments historyFileHasBeenSpecified]) {
    [theTopLevelSwarm writeHistoryToFile: [FearlusArguments getHistoryFile]];
  }
//  [theTopLevelSwarm drop]; // This causes a seg fault when dropping EZBar

  // theTopLevelSwarm has finished processing, so it's time to quit.
  return 0;
}
